#include <algorithm>
#include <iostream>
#include <string>
#include <vector>
#include <unistd.h>

using namespace std;

typedef vector<string> linearr;

/* Tehdnk turhia pakkauksia,
 * esim "nacube naball" -> "na{cube,ball}"
 * Jos kyll, -1. Muuten 0.
 */
static const int turhat = -1;
static const bool debug = false;
static const bool maysort = true;

class besttype
{
public:
	int gain;
	unsigned ind;
	unsigned pos;
	unsigned len;
public:
	besttype() { clear(); }
	void clear()
	{
		gain = turhat;
		ind = pos = len = 0;
	}
};

#include <map>

static linearr pack(const linearr &lines)
{
	if(debug)cerr << "Finding differences.\n";
	
	besttype best;
	
	for(unsigned a=0; a<lines.size(); ++a)
	{
		const string &s = lines[a];
		
		map<unsigned, pair<unsigned,unsigned> >matchlens;		
		for(unsigned b=a+1; b<lines.size(); ++b)
		{
			const string &ss = lines[b];
			const char *p = ss.c_str(), *q = s.c_str();
			unsigned maxlen = ss.size();
			if(maxlen > s.size())maxlen = s.size();
			unsigned alku = 0;
			while(alku<maxlen && *p == *q) { alku++; p++; q++; }
			
			p = s.c_str() + s.size();
			q = ss.c_str() + ss.size();
			unsigned loppu = 0;
			while(loppu<maxlen && *--p == *--q) loppu++;			
			
			matchlens[b] = pair<unsigned,unsigned> (alku, loppu);
		}

		/* Analyze the whole string... */
		for(unsigned pos=0; pos<=s.size(); ++pos)
		{
			/* Calculate longest continuous field at here */
			int len=s.size()-pos, safe=0;
			int state = 0;
			for(int tmp=0; tmp<=len; ++tmp)
			{
				/* Warning: pos+tmp may equal size() here */
				if(s[pos+tmp] == '{')
					++state;
				else if(s[pos+tmp] == '}')
					--state;
				
				if(!state)
					safe = tmp;
			}
			
			/* Then find substrings in everything else */
			for(len=safe; len>=0; --len)
			{
				unsigned loppusize = s.size() - (pos+len);
#if 0
				const string alku  = s.substr(0, pos);
				const string loppu = s.substr(pos+len);
				
				if(debug)
					cerr << "Trying '" << alku << "' +" << len << "+ '" << loppu << "'\n";
#endif
				
				unsigned commonlen = s.size() - len;
				
				unsigned count = 1;
				
				for(unsigned b=a+1; b<lines.size(); ++b)
				{
					const string &ss = lines[b];
					
                    if(matchlens[b].first < pos
                    || matchlens[b].second < loppusize)
                    {
                    	continue;
                    }
                    ++count;
				}
				
				if(count > 1)
				{
    				int oldlen = commonlen * count + (count-1);
    				int newlen = 1 + commonlen + (count-1) + 1;
    				
    				int gain = oldlen-newlen;
    				
    				if(debug)
	    				cerr << "Would gain " << gain << endl;
	    			
    				if(gain > best.gain)
    				{
    					best.gain = gain;
    					best.ind = a;
    					best.pos = pos;
    					best.len = len;
    				}
				}
			}
		}
	}
	if(debug)
		cerr << "Done...\n";

	if(debug)
		cerr << "Best gain: " << best.gain << endl;
	
	if(best.gain <= turhat)
	{
		return lines;
	}

    unsigned a   = best.ind;
    unsigned pos = best.pos;
    unsigned len = best.len;

    const string &s = lines[a];
    const string alku  = s.substr(0, pos);
    const string loppu = s.substr(pos+len);

    unsigned commonlen = alku.size() + loppu.size();

    linearr instead;
    linearr sels;
    
    for(unsigned b=lines.size();;)
    {
    	const string &ss = lines[--b];
    	
        if(ss.size() < commonlen
        || strncmp(ss.c_str(), alku.c_str(), alku.size())
        || strncmp(ss.c_str()+ss.size()-loppu.size(),
                               loppu.c_str(), loppu.size())
          )
        {
        	instead.push_back(ss);
        	continue;
        }
        
        sels.push_back(ss.substr(alku.size(), ss.size()-commonlen));
        
    	if(b==a)
    	{
    		while(b > 0)
    			instead.push_back(lines[--b]);
    		break;
    	}
    }
    
    if(maysort)
	    sort(sels.begin(), sels.end());
    sels = pack(sels);

    string replaced = alku;    
    if(sels.size() > 1)replaced += '{';

    for(unsigned a=0; a<sels.size(); ++a)
    {
    	if(a)replaced += ',';
    	replaced += sels[a];
    }
    
    if(sels.size() > 1)replaced += '}';        
    replaced += loppu;
    
    if(debug)
	    cerr << "Replaced things with '" << replaced << "'\n";
    
    instead.push_back(replaced);
    
    if(debug)
	    cerr << "Trying more...\n";
    
	if(maysort)
		sort(instead.begin(), instead.end());
    
    return pack(instead);
}

int main()
{
	if(isatty(0))
	{
		cout <<
			"Shellpack version "VERSION" - (C) 1992,2001 Bisqwit (http://iki.fi/bisqwit/)\n"
		    "Usage: shellpack < data.txt\n"
		    "\n"
		    "Compresses text lines with sh syntax.\n"
		    "\n"
		    "Example (/exec -msg bisqwit = ircII command, rest is shell commandline):\n"
		    "\n    /exec -msg bisqwit grep : .jnewsrc|sed 's/:.*//'|shellpack\n"
		    "\n"
		    "Produces compact output like \"sfnet.tiede.{bio,fysiikka}\".\n"
		    "\n"
		    "Note: This program is terribly slow with more than a few lines of data.\n"
		    "\n"
		;
		return 0;
	}

	vector <string> lines;

	for(;;)
	{
		string s;
		getline(cin, s);
		if(!cin.good())break;
		lines.push_back(s);
	}
	
	if(maysort)
		sort(lines.begin(), lines.end());
	
	lines = pack(lines);
	
    for(unsigned a=0; a<lines.size(); ++a)
    {
    	if(a)cout << ' ';
    	cout << lines[a];
    }
    cout << endl;
	
	return 0;
}
